package fullenv

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"gopkg.in/yaml.v2"
)

func TestApplyVars(t *testing.T) {
	for _, tt := range []struct {
		name     string
		initial  string
		vars     map[string]string
		expected string
	}{
		{"complex",
			`key: $V1
sub:
  key: $V2
list:
- key: $V3`,
			map[string]string{"V1": "value1", "V2": "value2", "V3": "value3"},
			`key: value1
sub:
  key: value2
list:
- key: value3
`,
		},
	} {
		t.Run(tt.name, func(t *testing.T) {
			var v RawSettings
			if err := yaml.Unmarshal([]byte(tt.initial), &v); err != nil {
				t.Fatal(err)
			}
			out, err := applyVars(v, tt.vars)
			if err != nil {
				t.Fatal(err)
			}
			b, err := yaml.Marshal(&out)
			if err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, tt.expected, string(b))
		})
	}
}
