// Copyright 2016 Orus.io. All rights reserved.

package storage

import (
	"time"

	"bitbucket.org/orus-io/go-xbus/api"
)

// ActorEventHandler is the type for callbacks
type ActorEventHandler func(api.Actor)

// NoopActorEventHandler is an ActorEventHandler that does nothing
func NoopActorEventHandler(api.Actor) {}

// ActorStorage is in charge of storing the actor definitions
// When multiple instances are running, CRUD operations be reflected from one
// to the others
type ActorStorage interface {
	Get(api.UUID) (api.Actor, bool)
	GetByName(name string) (api.Actor, bool)
	List() []api.Actor
	ListAccount(accountID api.UUID) []api.Actor
	ListKind(actorKind api.Actor_Kind) []api.Actor
	ListRole(actorKind api.Actor_Kind, role string) []api.Actor

	Create(api.Actor) (api.Actor, error)
	Delete(api.UUID)
	Update(api.Actor) (api.Actor, error)

	MarkUnresponsive(id api.UUID, decisionTime time.Time) error
	JustSeen(api.UUID) error
	SetOnline(api.UUID, bool) error

	SetEventHandlers(onActorUp, onActorDown ActorEventHandler)
}
