package mocks

import (
	"fmt"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// JobStorage is a mock for storage.Storage
type JobStorage struct {
	jobs         []*storage.Job
	index        map[storage.JobID]*storage.Job
	indexByActor map[api.UUID][]*storage.Job
}

// NewJobStorage ...
func NewJobStorage() *JobStorage {
	return &JobStorage{
		index:        make(map[storage.JobID]*storage.Job),
		indexByActor: make(map[api.UUID][]*storage.Job),
	}
}

func (s *JobStorage) add(job *storage.Job) {
	s.jobs = append(s.jobs, job)
	s.index[job.JobID] = job
	s.indexByActor[job.Node.ActorID] =
		append(s.indexByActor[job.Node.ActorID], job)
}

func (s *JobStorage) remove(jobID storage.JobID) {
	for i := range s.jobs {
		if s.jobs[i].JobID == jobID {
			s.jobs = append(s.jobs[:i], s.jobs[i+1:]...)
			break
		}
	}
	delete(s.index, jobID)

	if l, ok := s.indexByActor[jobID.Node.ActorID]; ok {
		for i := range l {
			if l[i].JobID == jobID {
				l = append(l[:i], l[i+1:]...)
				break
			}
		}
		s.indexByActor[jobID.Node.ActorID] = l
	}
}

// NewJob ...
func (s *JobStorage) NewJob(id storage.JobID, state storage.JobState) {
	s.add(&storage.Job{JobID: id, JobState: state})
	fmt.Println("Jobs", s.jobs)
}

// SetState ...
func (s *JobStorage) SetState(id storage.JobID, state storage.JobState) {
	job := storage.Job{JobID: id, JobState: state}
	s.remove(id)
	s.add(&job)
}

// GetState ...
func (s *JobStorage) GetState(id storage.JobID) storage.JobState {
	if job, ok := s.index[id]; ok {
		return job.JobState
	}
	panic("no such job")
}

// ListPendingJobByActor ...
func (s *JobStorage) ListPendingJobByActor(actorID api.UUID, after *storage.JobID, limit int) (jobs []storage.Job) {
	var found bool
	fmt.Println(s.jobs, actorID, after, s.indexByActor[actorID])
	for _, job := range s.indexByActor[actorID] {
		if !found && after != nil {
			if job.JobID == *after {
				found = true
			}
		} else if job.Status == api.ActorProcessingState_PENDING {
			jobs = append(jobs, *job)
		}
		if len(jobs) == limit {
			return
		}
	}
	return
}

// ListJobByProcess ...
func (s *JobStorage) ListJobByProcess(processID api.UUID) (jobs []storage.Job) {
	for _, job := range s.jobs {
		if job.ProcessID == processID && job.Status <= api.ActorProcessingState_RUNNING {
			jobs = append(jobs, *job)
		}
	}
	return
}

// ListRunningJobByEnvelopeID ...
func (s *JobStorage) ListRunningJobByEnvelopeID(envelopeID api.UUID) (jobs []storage.Job) {
	for _, job := range s.jobs {
		for _, id := range job.EnvelopeIDs {
			if id == envelopeID {
				jobs = append(jobs, *job)
				continue
			}
		}
	}
	return
}

// Purge ...
func (s *JobStorage) Purge(api.UUID) error {
	return nil
}
